/*******************************************************************
 * Copyright (C) 2000 SSI. 
 *
 * This is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this software. If not, write to the Free Software 
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 
 * 02111-1307, USA.
 ********************************************************************/
/**************************************************
 * RCSId: $Id$
 *
 * Calculate current observation point for polar satellites 
 * Project: SSI
 * Author: R. Alblas
 *
 * History: 
 * $Log$
 **************************************************/
/**************************************************
  Main functions:
    xy2lonlat(): translates (x,y) of picture into (lon,lat)
    calc_orbitconst()
 **************************************************/
/**************************************************
  Used formulas to calculate the satellite orbit:

    Used symbols:
      h=height sat
      Rearth=radius earth
      g0=gravity
      to=time 1 loop (in sec)
      lpd=loops per day = 24*3600/to
      orbit_nr=# orbits since reference time (=time Kepler data)
      dt=time elapsed since reference time
      orbit_angle=remainder orbit (orbit_nr - n*2pi)
      k2_n=deviation orbit north-south
      k2_w=deviation orbit west-east

    Common physics:
      to=2*pi*sqrt(Rearth/g0) * (1+h/Rearth)**(1.5)

          { [         to           ](2/3)     }
      h = { [ -------------------- ]      - 1 } * Rearth
          { [ 2*pi*sqrt(Rearth/g0) ]          }


    Satellite movement:
                                      
                     [  Rearth  ](3.5)            2
      k2_n = Const * [ -------- ]      * {2.5*(cos(incl)-1) + 2}
                     [ h+Rearth ]

      dh_n=dt*k2_n

                     [  Rearth  ](3.5)            
      k2_w = Const * [ -------- ]      * cos(incl)
                     [ h+Rearth ]

      dh_w=dt*k2_w

      orbit_nr=(lpd + decay*dt)*dt

      orbit_angle=(orbit_nr-int(orbit_nr))*2*pi + perigee + anomaly

      x1=sin(orbit_angle+dh_n*dt)*cos(incl)
      y1=cos(orbit_angle+dh_n*dt)
      z1=sin(orbit_angle+dh_n*dt)*sin(incl)

  Take into account sensor movement.
  beta=angle "sat -> earth-middle -> current observation point"
      x=x1*cos(beta)+sin(inclination)*sin(beta);
      y=y1*cos(beta);
      z=z1*cos(beta)-cos(inclination)*sin(beta);

   Translate to lon/lat
      sat_lon=atan(y/x)+raan-dh_w
      sat_lat=asin(z)

    Earth rotation:
      rem_e=(rot_year-int(rot_year))*2*pi
      earth_lon=rem_e + Earth_offsetangle
      earth_lat=0;

    Relative rotation. If beta=0 then this is the sub-satellite point.
      rel_lon=sat_lon-earth_lon
      rel_lat=sat_lat

 **************************************************/
#include <math.h>
#include <float.h>
#include <string.h>
#include "sattrack.h"
#include "xtrack_basefunc.h"

EPOINT sunpos,moonpos; 
/*************************************
 * Correct tm-struct and return #seconds since 1970.
 * Similar to mktime, except that dayligt saving is ignored.
 *************************************/
long mktime_ntz(struct tm *tm)
{
  long secs;
  int i;
  int days[]={31,28,31,30,31,30,31,31,30,31,30,31};

  while (tm->tm_sec >= 60) { tm->tm_min++; tm->tm_sec-=60; }
  while (tm->tm_sec <   0) { tm->tm_min--; tm->tm_sec+=60; }
  while (tm->tm_min >= 60) { tm->tm_hour++; tm->tm_min-=60; }
  while (tm->tm_min <   0) { tm->tm_hour--; tm->tm_min+=60; }
  while (tm->tm_hour>= 24) { tm->tm_mday++; tm->tm_hour-=24; }
  while (tm->tm_hour<   0) { tm->tm_mday--; tm->tm_hour+=24; }
  while (tm->tm_mon >= 12) { tm->tm_year++; tm->tm_mon-=12; }
  while (tm->tm_mon <   0) { tm->tm_year--; tm->tm_mon+=12; }
  if (tm->tm_year/4==tm->tm_year/4.) days[1]=29;

  while (tm->tm_mday>days[tm->tm_mon])
  {
    tm->tm_mday-=days[tm->tm_mon]; tm->tm_mon++;
    while (tm->tm_mon>= 12)  { tm->tm_year++; tm->tm_mon-=12; }
    if (tm->tm_year/4==tm->tm_year/4.) days[1]=29; else days[1]=28;
  }

  while (tm->tm_mday<=0)
  {
    tm->tm_mon--;
    tm->tm_mday+=days[tm->tm_mon];
    while (tm->tm_mon< 0)  { tm->tm_year--; tm->tm_mon+=12; }
    if (tm->tm_year/4==tm->tm_year/4.) days[1]=29; else days[1]=28;
  }

  secs=tm->tm_sec+60*(tm->tm_min+60*(tm->tm_hour+24*(tm->tm_mday-1)));
  for (i=0; i<tm->tm_mon; i++) secs+=(days[i]*24*3600);
  secs+=((tm->tm_year-70)*365*24*3600); 
  secs+=((tm->tm_year-70+1)/4)*24*3600; 
  tm->tm_yday=0;
  for (i=0; i<tm->tm_mon; i++) tm->tm_yday+=days[i];
  tm->tm_yday+=(tm->tm_mday-1);
  return secs;
}
/*************************************
 * translate days-in-year in tm struct into month and day
 *************************************/
void yday2mday_mon(struct tm *tmref)
{
  int i;
  int days[]={31,28,31,30,31,30,31,31,30,31,30,31};
  if (tmref->tm_year/4==tmref->tm_year/4.) days[1]=29;

  tmref->tm_mday=tmref->tm_yday+1;
  tmref->tm_mon=0; 
  for (i=0; i<12; i++)
  {
    if (tmref->tm_mday<=days[i]) break;
    tmref->tm_mon++;
    tmref->tm_mday-=days[i];
  }

}
/*************************************
 * Pre-calculate some orbit constants
 *************************************/
int calc_orbitconst(KEPLER *kepler,ORBIT *orbit)
{
  if (kepler->motion<0.1) return 0;

  orbit->loop_time=24.*3600./kepler->motion;
  orbit->height=(pow(orbit->loop_time/PIx2/sqrt(Rearth/G0),2./3.)-1)*Rearth;

  orbit->k2_n=Const * pow(Rearth/(Rearth+orbit->height),3.5) * 
                          (2.5*pow(cos(kepler->inclination),2.)-0.5);

  orbit->k2_w=Const * pow((Rearth/(Rearth+orbit->height)),3.5) * 
                                            cos(kepler->inclination);

  orbit->k2_n=D2R(orbit->k2_n);
  orbit->k2_w=D2R(orbit->k2_w);

/* Translate reference time into tm struct */
  orbit->ref_tm.tm_isdst=0;
  orbit->ref_tm.tm_year=kepler->epoch_year;
  orbit->ref_tm.tm_yday=kepler->epoch_day-1;
  yday2mday_mon(&orbit->ref_tm);
  orbit->ref_tm.tm_hour=  (kepler->epoch_day-1-orbit->ref_tm.tm_yday)*24;
  orbit->ref_tm.tm_min = ((kepler->epoch_day-1-orbit->ref_tm.tm_yday)*24-
                                               orbit->ref_tm.tm_hour)*60;
  orbit->ref_tm.tm_sec =(((kepler->epoch_day-1-orbit->ref_tm.tm_yday)*24-
                                               orbit->ref_tm.tm_hour)*60-
                                               orbit->ref_tm.tm_min )*60;

  orbit->ref_time=mktime_ntz(&orbit->ref_tm);
  orbit->ref_tm.tm_wday=0;
  orbit->ref_ms=0;
  struct tm momtm=mom_tm();
  orbit->data_age=(mktime_ntz(&momtm)-orbit->ref_time)/(24.*3600);

/* Set valid-flag */
  orbit->valid=TRUE;
  return 1;
}

/*************************************
 * Calculate position of satellite from 
 *   current time and reference time (ref-time is in KEPLER-data)
 * Result in pos_sat->x,y,z
 * raan and dh_w NOT taken into account. These have to be added to the lon.
 * Return: 0 if calculation has failed 
 *         1 if OK
 *************************************/
int calcpossat(struct tm *cur_tm, int cur_ms,  /* current time */
               KEPLER *kepler,                 /* kepler data */
               ORBIT *orbit,                   /* orbit data */
               EPOINT *pos_sat)                 /* position satellite */
{
  double dift;
  double loop_nr;
  float orbit_angle,dh_n,dh_w;

/* Determine time difference with reference in sec */
  dift=mktime_ntz(cur_tm)-orbit->ref_time;
  dift=dift+(float)(cur_ms-orbit->ref_ms)/1000.;

/* Translate difference into days */
  dift/=(24.*3600.);
/* Calculate # loops since ref-time. */
  loop_nr=(kepler->motion+kepler->decay_rate*dift)*dift; 
  if (ABS(loop_nr/10) >= FLT_MAX) return 0;     /* loop_nr too large -> error */

/* Calc orbit-angle in radians */
  orbit_angle=(loop_nr-(int)loop_nr)*PIx2;

/* Add starting point */
  orbit_angle=orbit_angle+kepler->perigee+kepler->anomaly;

/* Shift result to [0 ... pi*2] */
  while (orbit_angle<0)    orbit_angle+=PIx2;
  while (orbit_angle>PIx2) orbit_angle-=PIx2;

/* Determine corrections */
  dh_n=dift*orbit->k2_n;
  dh_w=dift*orbit->k2_w;

  dh_n=dh_n-(int)(dh_n/PIx2)*PIx2;
  dh_w=dh_w-(int)(dh_w/PIx2)*PIx2;
  orbit->dh_w=dh_w;

/* Translate to orthonormal coordinates */
  pos_sat->x=sin(orbit_angle+dh_n)*cos(kepler->inclination);
  pos_sat->y=cos(orbit_angle+dh_n);
  pos_sat->z=sin(orbit_angle+dh_n)*sin(kepler->inclination);

#define BEREKEN_MOM_HOOGTEn
#ifdef BEREKEN_MOM_HOOGTE
{
float loop_time=24.*3600./(kepler->motion+kepler->decay_rate*dift);
float height=(pow(loop_time/PIx2/sqrt(Rearth/G0),2./3.)-1)*Rearth;
printf("%f  %f  %f\n",height,kepler->decay_rate,(float)dift);
}
#endif
  return 1;
}


/*************************************
 * Calculate observation point of satellite.
 * Takes into account sensor movement.
 * Result in pos_sat->lon,lat
 *************************************/
void calc_satobs(KEPLER *kepler,ORBIT *orbit,
                 EPOINT *pos_sati,EPOINT *pos_sato,float beta)
{
/* No need to calc. if beta=0. Just to speed up calcs. */

  if (beta)
  {
    pos_sato->x=pos_sati->x*cos(beta)+sin(kepler->inclination)*sin(beta);
    pos_sato->y=pos_sati->y*cos(beta);
    pos_sato->z=pos_sati->z*cos(beta)-cos(kepler->inclination)*sin(beta);
  }
  else
  {
    pos_sato->x=pos_sati->x;
    pos_sato->y=pos_sati->y;
    pos_sato->z=pos_sati->z;
  }
}

/*************************************
 * Calculate position of earth since jan 1 
 *************************************/
void calcposearth(struct tm *cur_tm, int ms, EPOINT *pos_earth)
{
  double days,rot_earth;
/* days since jan 1 */
  days=cur_tm->tm_yday+1+cur_tm->tm_hour/24.+
                             cur_tm->tm_min/1440.+
                             (double)cur_tm->tm_sec/(double)86400.+
                             (float)ms/86400000.;

/* earth rotations since jan 1. Translate to 365.24 days/year. */
  rot_earth=days*LEN_LEAPYEAR/LEN_YEAR;

/* Translate into rotations. Ignore full rotations and add offset. */
  pos_earth->lon=PIx2*(rot_earth-(int)rot_earth)+D2R(Earth_offsetangle);

/* Translate to range [0 ... 2*pi] */
  while (pos_earth->lon>=PIx2) pos_earth->lon-=PIx2;

/* earth rotates in lon direction only */
  pos_earth->lat=0;
}

/*************************************
 * Calculate position of satellite relative to earth
 *************************************/
void calcposrel(KEPLER *kepler,ORBIT *orbit,
                EPOINT *pos_sat,EPOINT *pos_earth,EPOINT *pos_rel)
{
/* Translate to long/lat */

  pos_rel->lon=atan2(pos_sat->x,pos_sat->y)+
                          kepler->raan-orbit->dh_w-pos_earth->lon;
  pos_rel->lat=asin(pos_sat->z);

/* Shift to range [-pi..+pi] */  
  while (pos_rel->lon < -1*PI) pos_rel->lon+=PIx2;
  while (pos_rel->lon >    PI) pos_rel->lon-=PIx2;
  while (pos_rel->lat < -1*PI) pos_rel->lat+=PIx2;
  while (pos_rel->lat >    PI) pos_rel->lat-=PIx2;
}

/**************************************************
  Translation to elevation/azimuth:
    Rsat=h+Rearth;
    h1=sin(obs_lat)*sin(sat_lat) +cos(obs_lat)*cos(sat_lat)*cos(sat_lon-obs_lon);
    h=acos(h1);
    gamma=2*asin(sqrt((1-(h1))/2.) );

    dist=sqrt(Rearth*Rearth+Rsat*Rsat-2*Rearth*Rsat*cos(gamma));
    sat_elev=acos((Rearth-Rsat*cos(gamma))/dist)-pi/2

    sat_azim=acos((sin(sat_lat)-cos(h)*sin(obs_lat))/(cos(obs_lat)*sin(h)));

 **************************************************/

/*************************************
 * Calculate elevation/azimuth wrt observer position
 *************************************/
void calceleazim(EPOINT *pos_subsat,ORBIT *orbit,EPOINT *refpos,DIRECTION *satdir,gboolean x_at_disc)
{
  double gamma;
  double lon_o,lat_o,lon_s,lat_s;
  double dlon,h,h1;
  double dist,Rsat;
  lon_s=pos_subsat->lon; lat_s=pos_subsat->lat;
  lon_o=refpos->lon;     lat_o=refpos->lat;
  dlon=lon_s-lon_o;
  Rsat=orbit->height+Rearth;
  h1=sin(lat_o)*sin(lat_s) +cos(lat_o)*cos(lat_s)*cos(dlon);
  h=acos(h1);
  gamma=2*asin(sqrt((1-(h1))/2.) );

  dist=sqrt(Rearth*Rearth+Rsat*Rsat-2*Rearth*Rsat*cos(gamma));
  satdir->elev=acos((Rearth-Rsat*cos(gamma))/dist)-PI/2.;
  {
    double tmp=(sin(lat_s)-cos(h)*sin(lat_o))/(cos(lat_o)*sin(h));
    if (tmp>1.) tmp=1.; if (tmp<-1.) tmp=-1.;
    satdir->azim=acos(tmp);
  }
  if (satdir->azim<0) satdir->azim+=(PI*2);
  if (dlon<0) satdir->azim=PI*2-satdir->azim;
  if (x_at_disc)
  {
    satdir->y=atan2(cos(satdir->azim),tan(satdir->elev));
    satdir->x=asin(sin(satdir->azim)*cos(satdir->elev));
  }
  else
  {
    satdir->x=atan2(sin(satdir->azim),tan(satdir->elev));
    satdir->y=asin(cos(satdir->azim)*cos(satdir->elev));
  }
  satdir->x+=D2R(90.);
  satdir->y+=D2R(90.);
}

/*************************************
 * Translate sensor angle to angle position on earth.
 * (angle between line "sat <--> middle point earth" and "pos. earth <--> mp"
 * Needed: sat_height: height satellite
 *         x         : x-pos. (for HRPT: 0 ... 1023)
 *         xh        : half value of x (1023 for HRPT)
 *         alpha_max : max. angle sensor
 * If x=0  ==> alpha=alpha_max ==> beta=beta_max
 * if x=xh ==> alpha=0 ==> beta=0        
 *************************************/
float x2earthangle(float sat_height,            /* height sat */
                   SCANDIR scan_hor,            /* scan direction sensor */
                   int x,int xh,                 
                   float alpha_max)
{
  float alpha,beta,tmp1,tmp2;

/* angle sensor in satellite */
  alpha=(xh-x)*alpha_max/xh;
  if (scan_hor==W_E) alpha*=-1;

/* translate into radius angle */
  tmp1=(sat_height+Rearth)/Rearth;
  tmp2=(tmp1*tmp1*sin(alpha)*sin(alpha));
  beta=acos(tmp1*sin(alpha)*sin(alpha)+
                 cos(alpha)*sqrt(1-tmp2));
  if (alpha<0) beta*=-1;
  return beta;
}
void calc_x2satobs(KEPLER *kepler,ORBIT *orbit,
                   EPOINT *pos_sati,EPOINT *pos_sato,
                   int width_original,float max_sens_angle,
                   int x)
{
  float beta;
  beta=x2earthangle(orbit->height,E_W,
                    x,width_original/2,max_sens_angle);
  calc_satobs(kepler,orbit,pos_sati,pos_sato,beta);
//printf("beta=%f   %f  %f  %f\n",beta,pos_sati->x,pos_sati->y,pos_sati->z);
}

#ifdef XXX
pri_kep(KEPLER kepler)
{
printf("ey=%d  ed=%f  dr=%f  inc=%f  ra=%f  ecc=%f  per=%f  ano=%f  mot=%f\n",
kepler.epoch_year,kepler.epoch_day,kepler.decay_rate,kepler.inclination,
kepler.raan,kepler.eccentricity,kepler.perigee,kepler.anomaly,kepler.motion);
}
#endif

/********************************************************
 * Calc. sun
 ********************************************************/
  //# Astronomer's almanach time is the number of 
  //# days since (noon, 1 January 2000)
static float astronomersAlmanacTime(struct tm tm)
{
  struct tm tm1;
  memset(&tm1,0,sizeof(tm1));
  tm1.tm_mday=1;
  tm1.tm_mon=0;
  tm1.tm_year=2000-1900;
  tm1.tm_hour=12;
  long tx=mktime_ntz(&tm);
  long ty=mktime_ntz(&tm1);
  float dt=difftime(tx,ty);
  return dt/(24.*3600.);
}

static float hourOfDay(struct tm tm)
{
  float h=tm.tm_hour+tm.tm_min/60.+tm.tm_sec/3600.;
  return h;
}

static float meanLongitudeDegrees(float time)
{
  float ret;
  ret=(280.460 + 0.9856474 * time);
  while (ret>360) ret-=360;
  while (ret<0) ret+=360;
  return ret;
}

static float meanAnomalyRadians(float time)
{
  float ret;
  ret=357.528 + 0.9856003 * time;
  while (ret>360) ret-=360;
  while (ret<0) ret+=360;
  return D2R(ret);
}

static float eclipticLongitudeRadians(float mnlong, float mnanom)
{
  float ret;
  ret=mnlong + 1.915 * sin(mnanom) + 0.020 * sin(2 * mnanom);
  while (ret>360) ret-=360;
  while (ret<0) ret+=360;
  return D2R(ret);
}

static float eclipticObliquityRadians(float time)
{
  return D2R(23.439 - 0.0000004 * time);
}

static float rightAscensionRadian(float oblqec, float eclong)
{
  float num,den,ra;
  num=cos(oblqec) * sin(eclong);
  den=cos(eclong);
  ra=atan2(num,den);
  return ra;
}

static float rightDeclinationRadians(float oblqec, float eclong)
{
  return asin(sin(oblqec) * sin(eclong));
}


static float greenwichMeanSiderealTimeHours(float time, float hour)
{
  float ret;
  ret=(6.697375 + 0.0657098242 * time + hour);
  while (ret>24) ret-=24;
  while (ret<0) ret+=24;
  return ret;
}

static float localMeanSiderealTimeRadians(float gmst, float lon)
{
  float ret=(gmst + lon / 15.);
  while (ret>24) ret-=24;
  while (ret<0) ret+=24;
  return D2R(15.*ret);
}

static float hourAngleRadians(float lmst, float ra)
{
  float ret;
  ret=lmst - ra + PI;
  while (ret>2.*PI) ret-=2.*PI;
  return ret-PI;
}


static EPOINT calc_sun(struct tm tm)
{
  EPOINT zon;
  float time,fhour;
  float lon=0.;
  float mnlong,mnanom,eclong,oblqec;
  float ra,dec,gmst,lmst;
  float ha;
  time=astronomersAlmanacTime(tm);
  fhour = hourOfDay(tm);

  //# Ecliptic coordinates  
  mnlong = meanLongitudeDegrees(time);  
  mnanom = meanAnomalyRadians(time);
  eclong = eclipticLongitudeRadians(mnlong, mnanom);   
  oblqec = eclipticObliquityRadians(time);

  //# Celestial coordinates
  ra =rightAscensionRadian(oblqec, eclong);
  dec = rightDeclinationRadians(oblqec, eclong);

  //# Local coordinates
  gmst = greenwichMeanSiderealTimeHours(time, fhour) ; 
  lmst = localMeanSiderealTimeRadians(gmst, lon);

  //# Hour angle
  ha = hourAngleRadians(lmst, ra);

  zon.lon=-1.*ha;
  zon.lat=dec;
  return zon;
}

 int T45AD[] = {0, 2, 2, 0, 0, 0, 2, 2, 2, 2,
                         0, 1, 0, 2, 0, 0, 4, 0, 4, 2,
                         2, 1, 1, 2, 2, 4, 2, 0, 2, 2,
                         1, 2, 0, 0, 2, 2, 2, 4, 0, 3,
                         2, 4, 0, 2, 2, 2, 4, 0, 4, 1,
                         2, 0, 1, 3, 4, 2, 0, 1, 2, 2};

  int  T45AM[] = {0,  0,  0,  0,  1,  0,  0, -1,  0, -1,
                         1,  0,  1,  0,  0,  0,  0,  0,  0,  1,
                         1,  0,  1, -1,  0,  0,  0,  1,  0, -1,
                         0, -2,  1,  2, -2,  0,  0, -1,  0,  0,
                         1, -1,  2,  2,  1, -1,  0,  0, -1,  0,
                         1,  0,  1,  0,  0, -1,  2,  1,  0,  0};

   int T45AMP[] = { 1, -1,  0,  2,  0,  0, -2, -1,  1,  0,
                          -1,  0,  1,  0,  1,  1, -1,  3, -2, -1,
                           0, -1,  0,  1,  2,  0, -3, -2, -1, -2,
                           1,  0,  2,  0, -1,  1,  0, -1,  2, -1,
                           1, -2, -1, -1, -2,  0,  1,  4,  0, -2,
                           0,  2,  1, -2, -3,  2,  1, -1,  3, -1};

   int T45AF[]  = { 0,  0,  0,  0,  0,  2,  0,  0,  0,  0,
                           0,  0,  0, -2,  2, -2,  0,  0,  0,  0,
                           0,  0,  0,  0,  0,  0,  0,  0,  2,  0,
                           0,  0,  0,  0,  0, -2,  2,  0,  2,  0,
                           0,  0,  0,  0,  0, -2,  0,  0,  0,  0,
                          -2, -2,  0,  0,  0,  0,  0,  0,  0, -2};

   int T45AL[] = {6288774, 1274027, 658314, 213618, -185116,
                         -114332,   58793,  57066,  53322,   45758,
                          -40923,  -34720, -30383,  15327,  -12528,
                           10980,   10675,  10034,   8548,   -7888,
                           -6766,   -5163,   4987,   4036,    3994,
                            3861,    3665,  -2689,  -2602,    2390,
                           -2348,    2236,  -2120,  -2069,    2048,
                           -1773,   -1595,   1215,  -1110,    -892,
                            -810,     759,   -713,   -700,     691,
                             596,     549,    537,    520,    -487,
                            -399,    -381,    351,   -340,     330,
                             327,    -323,    299,    294,       0};

   int T45AR[] = {-20905355, -3699111, -2955968, -569925,   48888,
                             -3149,   246158,  -152138, -170733, -204586,
                           -129620,   108743,   104755,   10321,       0,
                             79661,   -34782,   -23210,  -21636,   24208,
                             30824,    -8379,   -16675,  -12831,  -10445,
                            -11650,    14403,    -7003,       0,   10056,
                              6322,    -9884,     5751,       0,   -4950,
                              4130,        0,    -3958,       0,    3258,
                              2616,    -1897,    -2117,    2354,       0,
                                 0,    -1423,    -1117,   -1571,   -1739,
                                 0,    -4421,        0,       0,       0,
                                 0,     1165,        0,       0,    8752};

// Meeus table 45B latitude of the moon

   int T45BD[] = {0, 0, 0, 2, 2, 2, 2, 0, 2, 0,
                         2, 2, 2, 2, 2, 2, 2, 0, 4, 0,
                         0, 0, 1, 0, 0, 0, 1, 0, 4, 4,
                         0, 4, 2, 2, 2, 2, 0, 2, 2, 2,
                         2, 4, 2, 2, 0, 2, 1, 1, 0, 2,
                         1, 2, 0, 4, 4, 1, 4, 1, 4, 2};

   int T45BM[] = { 0,  0,  0,  0,  0,  0,  0, 0,  0,  0,
                         -1,  0,  0,  1, -1, -1, -1, 1,  0,  1,
                          0,  1,  0,  1,  1,  1,  0, 0,  0,  0,
                          0,  0,  0,  0, -1,  0,  0, 0,  0,  1,
                          1,  0, -1, -2,  0,  1,  1, 1,  1,  1,
                          0, -1,  1,  0, -1,  0,  0, 0, -1, -2};

   int T45BMP[] = {0,  1, 1,  0, -1, -1,  0,  2,  1,  2,
                          0, -2, 1,  0, -1,  0, -1, -1, -1,  0,
                          0, -1, 0,  1,  1,  0,  0,  3,  0, -1,
                          1, -2, 0,  2,  1, -2,  3,  2, -3, -1,
                          0,  0, 1,  0,  1,  1,  0,  0, -2, -1,
                          1, -2, 2, -2, -1,  1,  1, -1,  0,  0};

   int T45BF[] = { 1,  1, -1, -1,  1, -1,  1,  1, -1, -1,
                         -1, -1,  1, -1,  1,  1, -1, -1, -1,  1,
                          3,  1,  1,  1, -1, -1, -1,  1, -1,  1,
                         -3,  1, -3, -1, -1,  1, -1,  1, -1,  1,
                          1,  1,  1, -1,  3, -1, -1,  1, -1, -1,
                          1, -1,  1, -1, -1, -1, -1, -1, -1,  1};

   int T45BL[] = {5128122, 280602, 277693, 173237, 55413,
                           46271,  32573,  17198,   9266,  8822,
                            8216,   4324,   4200,  -3359,  2463,
                            2211,   2065,  -1870,   1828, -1794,
                           -1749,  -1565,  -1491,  -1475, -1410,
                           -1344,  -1335,   1107,   1021,   833,
                             777,    671,    607,    596,   491,
                            -451,    439,    422,    421,  -366,
                            -351,    331,    315,    302,  -283,
                            -229,    223,    223,   -220,  -220,
                            -185,    181,   -177,    176,   166,
                            -164,    132,   -119,    115,   107};

float rev(angle){return angle-(int)(angle/360.0)*360.0;}
float sind(angle){return sin((angle*PI)/180.0);}
float cosd(angle){return cos((angle*PI)/180.0);}
float tand(angle){return tan((angle*PI)/180.0);}
float asind(c){return (180.0/PI)*asin(c);}
//float acosd(c){return (180.0/PI)*acos(c);}
float atan2d(y,x){return (180.0/PI)*atan2(y,x)-180.0*(x<0);}

//float rev(a) {return a;}
static EPOINT calc_moon1(struct tm tm)
{
  EPOINT maan;
  float jd;
  jd=astronomersAlmanacTime(tm);
printf("%f\n",jd);
 float T=jd;
 float  T2=T*T;
 float  T3=T2*T;
 float  T4=T3*T;
  // Moons mean longitude L'
  float LP=218.3164477+481267.88123421*T-0.0015786*T2+T3/538841.0-T4/65194000.0;
  // Moons mean elongation
  float D=297.8501921+445267.1114034*T-0.0018819*T2+T3/545868.0-T4/113065000.0;
  // Suns mean anomaly
 float  M=357.5291092+35999.0502909*T-0.0001536*T2+T3/24490000.0;
  // Moons mean anomaly M'
 float  MP=134.9633964+477198.8675055*T+0.0087414*T2+T3/69699.0-T4/14712000.0;
  // Moons argument of latitude
 float  F=93.2720950+483202.0175233*T-0.0036539*T2-T3/3526000.0+T4/863310000.0;

  // Additional arguments
  float A1=119.75+131.849*T;
  float A2=53.09+479264.290*T;
  float A3=313.45+481266.484*T;
  float E=1-0.002516*T-0.0000074*T2;
 float  E2=E*E;
  // Sums of periodic terms from table 45.A and 45.B
  float Sl=0.0;
  float Sr=0.0;
  int i;
  for ( i=0; i<60; i++) {
    float Eterm=1;
    if (ABS(T45AM[i])==1) Eterm=E;
    if (ABS(T45AM[i])==2) Eterm=E2;
    Sl+=T45AL[i]*Eterm*sind(rev(T45AD[i]*D+T45AM[i]*M+T45AMP[i]*MP+T45AF[i]*F));
    Sr+=T45AR[i]*Eterm*cosd(rev(T45AD[i]*D+T45AM[i]*M+T45AMP[i]*MP+T45AF[i]*F));
  }
  float Sb=0.0;
  for ( i=0; i<60; i++) {
    float Eterm=1;
    if (ABS(T45BM[i])==1) Eterm=E;
    if (ABS(T45BM[i])==2) Eterm=E2;
    Sb+=T45BL[i]*Eterm*sind(rev(T45BD[i]*D+T45BM[i]*M+T45BMP[i]*MP+T45BF[i]*F));
  }
  // Additional additive terms
  Sl=Sl+3958*sind(rev(A1))+1962*sind(rev(LP-F))+318*sind(rev(A2));
  Sb=Sb-2235*sind(rev(LP))+382*sind(rev(A3))+175*sind(rev(A1-F))+
     175*sind(rev(A1+F))+127*sind(rev(LP-MP))-115*sind(rev(LP+MP));
  // geocentric longitude, latitude and distance
  float mglong=rev(LP+Sl/1000000.0);
  float mglat=rev(Sb/1000000.0);
  if (mglat > 180.0) mglat=mglat-360;
  float mr=(385000.56+Sr/1000.0);
  // Obliquity of Ecliptic
  float obl=23.4393-3.563E-9*(jd-2451543.5);
  // RA and dec
  float ra=rev(atan2d(sind(mglong)*cosd(obl)-tand(mglat)*sin(obl),
                    cosd(mglong)))/15.0;
  float dec=rev(asind(sind(mglat)*cosd(obl)+cosd(mglat)*sind(obl)*sind(mglong)));
  if (dec > 180.0) dec=dec-360;
  maan.lon = D2R(dec);
  maan.lat = D2R(ra);
printf("%f  %f\n",R2D(dec),R2D(ra));
  return maan; //  new Array(ra,dec,mr);
}



static EPOINT calc_moon(struct tm tm)
{
  EPOINT maan;
  float d,fhour;
  d=astronomersAlmanacTime(tm);
  float N = (125.1228 - 0.0529538083 * d) / 180 * PI;
  float i = 5.1454;
  float w = (318.0634 + 0.1643573223 * d) / 180 * PI;
  float e = 0.054900; //eccentricity
  float M = (115.3654 + 13.0649929509 * d) / 180.0f * PI ; //mean anomaly

  float E = M + e * sin(M) * ( 1.0 + e * cos(M) );
  E = E - ( E - e * sin(E) - M ) / ( 1 - e * cos(E) );

  float xv = ( cos(E) - e );
  float yv = ( sqrt(1.0 - e*e) * sin(E) );

  float v = atan2( yv, xv );

  float xh = ( cos(N) * cos(v+w) - sin(N) * sin(v+w) * cos(i) );
  float yh = ( sin(N) * cos(v+w) + cos(N) * sin(v+w) * cos(i) );
  float zh = ( sin(v+w) * sin(i) );

  maan.lon = atan2( yh, xh ); //longitude
  maan.lat = atan2( zh, sqrt(xh*xh+yh*yh) ); //latitude
  return maan;
}

#ifdef ZONSIMPEL
// lente: 31+28+22=81
// zomer: 81+9+30+31+22=173
// herfst:173+9+31+31+22=266
// winter: 266+8+31+30+22=357
#define LAT_KEERKR 23.439
static EPOINT calc_zon_simpel(struct tm ti)
{
  struct tm t=ti;
  EPOINT zon;
  int tjaar;
  int tdag;
  float zlon;
  float zlat;
  t.tm_year=70;
  tjaar=mktime_ntz(&t);                        // sec in jaar
  tdag=t.tm_hour*3600+t.tm_min*60+t.tm_sec;   // sec in dag
  zlon=((12.-(float)(tdag)/3600.))*90./6.;
  zlat=(float)(tjaar-(173*24*3600))/(float)((173*24*3600));
  zlat=LAT_KEERKR*cos(D2R(zlat*180.));

  zon.lon=D2R(zlon);
  zon.lat=D2R(zlat);
//printf("%f  %f\n",R2D(zon.lon),R2D(zon.lat));
  return zon;
}
#endif

/* Calc. position of satellite at time 'cur_tm' */
void calc_onepos(SAT *sat,struct tm cur_tm,int cur_ms,
                 EPOINT *pos_earth,EPOINT *pos_sat,EPOINT *pos_subsat)
{
  struct tm cur_gmtm;
  cur_gmtm=cur_tm;
  if (sat->type==satellite)
  {
    calcposearth(&cur_gmtm,cur_ms,pos_earth);
    calcpossat(&cur_gmtm,cur_ms,&sat->kepler,&sat->orbit,pos_sat);
    calc_satobs(&sat->kepler,&sat->orbit,pos_sat,pos_sat,0);
    calcposrel(&sat->kepler,&sat->orbit,pos_sat,pos_earth,pos_subsat);
  }
  else if (sat->type==sun)
  {
    *pos_subsat=calc_sun(cur_tm);
    sunpos.lon=pos_subsat->lon;
    sunpos.lat=pos_subsat->lat;
  }
  else
  {
    *pos_subsat=calc_moon(cur_tm);
    moonpos.lon=pos_subsat->lon;
    moonpos.lat=pos_subsat->lat;
  }
}

void calc_satpos(struct tm cur_tm,EPOINT *pos_subsat,DIRECTION *satdir,EPOINT *refpos,SAT *sat,gboolean x_at_disc)
{
  EPOINT pos_sat;
  EPOINT pos_earth;
  calc_onepos(sat,cur_tm,0,&pos_earth,&pos_sat,pos_subsat);
  calceleazim(pos_subsat,&sat->orbit,refpos,satdir,x_at_disc);
}

